#pragma once

#include <stdint.h>

#include "shared.h"
#include "cpu.h"
#include "instructions.h"

#define DEFINE_MODRM_INSTR_READ16(name, fun) \
    void name ## _mem(int32_t addr, int32_t r) { int32_t ___ = safe_read16(addr); fun; } \
    void name ## _reg(int32_t r1, int32_t r) { int32_t ___ = read_reg16(r1); fun; }

#define DEFINE_MODRM_INSTR_READ32(name, fun) \
    void name ## _mem(int32_t addr, int32_t r) { int32_t ___ = safe_read32s(addr); fun; } \
    void name ## _reg(int32_t r1, int32_t r) { int32_t ___ = read_reg32(r1); fun; }


#define DEFINE_MODRM_INSTR_IMM_READ_WRITE_16(name, fun) \
    void name ## _mem(int32_t addr, int32_t r, int32_t imm) { SAFE_READ_WRITE16(addr, fun) } \
    void name ## _reg(int32_t r1, int32_t r, int32_t imm) { int32_t ___ = read_reg16(r1); write_reg16(r1, fun); }

#define DEFINE_MODRM_INSTR_IMM_READ_WRITE_32(name, fun) \
    void name ## _mem(int32_t addr, int32_t r, int32_t imm) { SAFE_READ_WRITE32(addr, fun) } \
    void name ## _reg(int32_t r1, int32_t r, int32_t imm) { int32_t ___ = read_reg32(r1); write_reg32(r1, fun); }


#define DEFINE_MODRM_INSTR_READ_WRITE_8(name, fun) \
    void name ## _mem(int32_t addr, int32_t r) { SAFE_READ_WRITE8(addr, fun) } \
    void name ## _reg(int32_t r1, int32_t r) { int32_t ___ = read_reg8(r1); write_reg8(r1, fun); }

#define DEFINE_MODRM_INSTR_READ_WRITE_16(name, fun) \
    void name ## _mem(int32_t addr, int32_t r) { SAFE_READ_WRITE16(addr, fun) } \
    void name ## _reg(int32_t r1, int32_t r) { int32_t ___ = read_reg16(r1); write_reg16(r1, fun); }

#define DEFINE_MODRM_INSTR_READ_WRITE_32(name, fun) \
    void name ## _mem(int32_t addr, int32_t r) { SAFE_READ_WRITE32(addr, fun) } \
    void name ## _reg(int32_t r1, int32_t r) { int32_t ___ = read_reg32(r1); write_reg32(r1, fun); }


#define DEFINE_SSE_SPLIT(name, mem_fn, reg_fn) \
    void name ## _reg(int32_t r1, int32_t r2) { name(reg_fn(r1), r2); } \
    void name ## _mem(int32_t addr, int32_t r) { name(mem_fn(addr), r); } \

#define DEFINE_SSE_SPLIT_IMM(name, mem_fn, reg_fn) \
    void name ## _reg(int32_t r1, int32_t r2, int32_t imm) { name(reg_fn(r1), r2, imm); } \
    void name ## _mem(int32_t addr, int32_t r, int32_t imm) { name(mem_fn(addr), r, imm); } \

#define DEFINE_SSE_SPLIT_WRITE(name, mem_fn, reg_fn) \
    void name ## _reg(int32_t r1, int32_t r2) { reg_fn(r1, name(r2)); } \
    void name ## _mem(int32_t addr, int32_t r) { mem_fn(addr, name(r)); } \

void instr_0F00_0_mem(int32_t addr);
void instr_0F00_0_reg(int32_t r);
void instr_0F00_1_mem(int32_t addr);
void instr_0F00_1_reg(int32_t r);
void instr_0F00_2_mem(int32_t addr);
void instr_0F00_2_reg(int32_t r);
void instr_0F00_3_mem(int32_t addr);
void instr_0F00_3_reg(int32_t r);
void instr_0F00_4_mem(int32_t addr);
void instr_0F00_4_reg(int32_t r);
void instr_0F00_5_mem(int32_t addr);
void instr_0F00_5_reg(int32_t r);
void instr_0F01_0_reg(int32_t r);
void instr_0F01_0_mem(int32_t addr);
void instr_0F01_1_reg(int32_t r);
void instr_0F01_1_mem(int32_t addr);
void instr_0F01_2_reg(int32_t r);
void instr_0F01_2_mem(int32_t addr);
void instr_0F01_3_reg(int32_t r);
void instr_0F01_3_mem(int32_t addr);
void instr_0F01_4_reg(int32_t r);
void instr_0F01_4_mem(int32_t addr);
void lmsw(int32_t new_cr0);
void instr_0F01_6_reg(int32_t r);
void instr_0F01_6_mem(int32_t addr);
void instr_0F01_7_reg(int32_t r);
void instr_0F01_7_mem(int32_t addr);
void instr16_0F02_mem(int32_t addr, int32_t r);
void instr16_0F02_reg(int32_t r1, int32_t r);
void instr32_0F02_mem(int32_t addr, int32_t r);
void instr32_0F02_reg(int32_t r1, int32_t r);
void instr16_0F03_mem(int32_t addr, int32_t r);
void instr16_0F03_reg(int32_t r1, int32_t r);
void instr32_0F03_mem(int32_t addr, int32_t r);
void instr32_0F03_reg(int32_t r1, int32_t r);
void instr_0F04(void);
void instr_0F05(void);
void instr_0F06(void);
void instr_0F07(void);
void instr_0F08(void);
void instr_0F09(void);
void instr_0F0A(void);
void instr_0F0B(void);
void instr_0F0C(void);
void instr_0F0D(void);
void instr_0F0E(void);
void instr_0F0F(void);
void instr_0F10(union reg128 source, int32_t r);
void instr_0F10_reg(int32_t r1, int32_t r2);
void instr_0F10_mem(int32_t addr, int32_t r);
void instr_F30F10_reg(int32_t r1, int32_t r2);
void instr_F30F10_mem(int32_t addr, int32_t r);
void instr_660F10(union reg128 source, int32_t r);
void instr_660F10_reg(int32_t r1, int32_t r2);
void instr_660F10_mem(int32_t addr, int32_t r);
void instr_F20F10_reg(int32_t r1, int32_t r2);
void instr_F20F10_mem(int32_t addr, int32_t r);
void instr_0F11_reg(int32_t r1, int32_t r2);
void instr_0F11_mem(int32_t addr, int32_t r);
void instr_F30F11_reg(int32_t rm_dest, int32_t reg_src);
void instr_F30F11_mem(int32_t addr, int32_t r);
void instr_660F11_reg(int32_t r1, int32_t r2);
void instr_660F11_mem(int32_t addr, int32_t r);
void instr_F20F11_reg(int32_t r1, int32_t r2);
void instr_F20F11_mem(int32_t addr, int32_t r);
void instr_0F12_mem(int32_t addr, int32_t r);
void instr_0F12_reg(int32_t r1, int32_t r2);
void instr_660F12_reg(int32_t r1, int32_t r);
void instr_660F12_mem(int32_t addr, int32_t r);
void instr_F20F12_mem(int32_t addr, int32_t r);
void instr_F20F12_reg(int32_t r1, int32_t r2);
void instr_F30F12_mem(int32_t addr, int32_t r);
void instr_F30F12_reg(int32_t r1, int32_t r2);
void instr_0F13_mem(int32_t addr, int32_t r);
void instr_0F13_reg(int32_t r1, int32_t r2);
void instr_660F13_reg(int32_t r1, int32_t r);
void instr_660F13_mem(int32_t addr, int32_t r);
void instr_0F14(union reg64 source, int32_t r);
void instr_0F14_reg(int32_t r1, int32_t r2);
void instr_0F14_mem(int32_t addr, int32_t r);
void instr_660F14(union reg64 source, int32_t r);
void instr_660F14_reg(int32_t r1, int32_t r2);
void instr_660F14_mem(int32_t addr, int32_t r);
void instr_0F15(union reg128 source, int32_t r);
void instr_0F15_reg(int32_t r1, int32_t r2);
void instr_0F15_mem(int32_t addr, int32_t r);
void instr_660F15(union reg128 source, int32_t r);
void instr_660F15_reg(int32_t r1, int32_t r2);
void instr_660F15_mem(int32_t addr, int32_t r);
void instr_0F16_mem(int32_t addr, int32_t r);
void instr_0F16_reg(int32_t r1, int32_t r2);
void instr_660F16_mem(int32_t addr, int32_t r);
void instr_660F16_reg(int32_t r1, int32_t r2);
void instr_0F17_mem(int32_t addr, int32_t r);
void instr_0F17_reg(int32_t r1, int32_t r2);
void instr_660F17_mem(int32_t addr, int32_t r);
void instr_660F17_reg(int32_t r1, int32_t r2);
void instr_0F18_reg(int32_t r1, int32_t r2);
void instr_0F18_mem(int32_t addr, int32_t r);
void instr_0F19(void);
void instr_0F1A(void);
void instr_0F1B(void);
void instr_0F1C(void);
void instr_0F1D(void);
void instr_0F1E(void);
void instr_0F1F_reg(int32_t r1, int32_t r2);
void instr_0F1F_mem(int32_t addr, int32_t r);
void instr_0F20(int32_t r, int32_t creg);
void instr_0F21(int32_t r, int32_t dreg_index);
void instr_0F22(int32_t r, int32_t creg);
void instr_0F23(int32_t r, int32_t dreg_index);
void instr_0F24(void);
void instr_0F25(void);
void instr_0F26(void);
void instr_0F27(void);
void instr_0F28(union reg128 source, int32_t r);
void instr_0F28_reg(int32_t r1, int32_t r2);
void instr_0F28_mem(int32_t addr, int32_t r);
void instr_660F28(union reg128 source, int32_t r);
void instr_660F28_reg(int32_t r1, int32_t r2);
void instr_660F28_mem(int32_t addr, int32_t r);
void instr_0F29_mem(int32_t addr, int32_t r);
void instr_0F29_reg(int32_t r1, int32_t r2);
void instr_660F29_mem(int32_t addr, int32_t r);
void instr_660F29_reg(int32_t r1, int32_t r2);
void instr_0F2A(void);
void instr_0F2B_reg(int32_t r1, int32_t r2);
void instr_0F2B_mem(int32_t addr, int32_t r);
void instr_660F2B_reg(int32_t r1, int32_t r2);
void instr_660F2B_mem(int32_t addr, int32_t r);
void instr_0F2C_mem(int32_t addr, int32_t r);
void instr_0F2C_reg(int32_t r1, int32_t r2);
void instr_660F2C_mem(int32_t addr, int32_t r);
void instr_660F2C_reg(int32_t r1, int32_t r2);
void instr_F20F2C(union reg64 source, int32_t r);
void instr_F20F2C_reg(int32_t r1, int32_t r2);
void instr_F20F2C_mem(int32_t addr, int32_t r);
void instr_F30F2C_mem(int32_t addr, int32_t r);
void instr_F30F2C_reg(int32_t r1, int32_t r2);
void instr_0F2D(void);
void instr_0F2E(void);
void instr_0F2F(void);
void instr_0F30(void);
void instr_0F31(void);
void instr_0F32(void);
void instr_0F33(void);
void instr_0F34(void);
void instr_0F35(void);
void instr_0F36(void);
void instr_0F37(void);
void instr_0F38(void);
void instr_0F39(void);
void instr_0F3A(void);
void instr_0F3B(void);
void instr_0F3C(void);
void instr_0F3D(void);
void instr_0F3E(void);
void instr_0F3F(void);
void instr16_0F40_mem(int32_t addr, int32_t r);
void instr16_0F40_reg(int32_t r1, int32_t r);
void instr32_0F40_mem(int32_t addr, int32_t r);
void instr32_0F40_reg(int32_t r1, int32_t r);
void instr16_0F41_mem(int32_t addr, int32_t r);
void instr16_0F41_reg(int32_t r1, int32_t r);
void instr32_0F41_mem(int32_t addr, int32_t r);
void instr32_0F41_reg(int32_t r1, int32_t r);
void instr16_0F42_mem(int32_t addr, int32_t r);
void instr16_0F42_reg(int32_t r1, int32_t r);
void instr32_0F42_mem(int32_t addr, int32_t r);
void instr32_0F42_reg(int32_t r1, int32_t r);
void instr16_0F43_mem(int32_t addr, int32_t r);
void instr16_0F43_reg(int32_t r1, int32_t r);
void instr32_0F43_mem(int32_t addr, int32_t r);
void instr32_0F43_reg(int32_t r1, int32_t r);
void instr16_0F44_mem(int32_t addr, int32_t r);
void instr16_0F44_reg(int32_t r1, int32_t r);
void instr32_0F44_mem(int32_t addr, int32_t r);
void instr32_0F44_reg(int32_t r1, int32_t r);
void instr16_0F45_mem(int32_t addr, int32_t r);
void instr16_0F45_reg(int32_t r1, int32_t r);
void instr32_0F45_mem(int32_t addr, int32_t r);
void instr32_0F45_reg(int32_t r1, int32_t r);
void instr16_0F46_mem(int32_t addr, int32_t r);
void instr16_0F46_reg(int32_t r1, int32_t r);
void instr32_0F46_mem(int32_t addr, int32_t r);
void instr32_0F46_reg(int32_t r1, int32_t r);
void instr16_0F47_mem(int32_t addr, int32_t r);
void instr16_0F47_reg(int32_t r1, int32_t r);
void instr32_0F47_mem(int32_t addr, int32_t r);
void instr32_0F47_reg(int32_t r1, int32_t r);
void instr16_0F48_mem(int32_t addr, int32_t r);
void instr16_0F48_reg(int32_t r1, int32_t r);
void instr32_0F48_mem(int32_t addr, int32_t r);
void instr32_0F48_reg(int32_t r1, int32_t r);
void instr16_0F49_mem(int32_t addr, int32_t r);
void instr16_0F49_reg(int32_t r1, int32_t r);
void instr32_0F49_mem(int32_t addr, int32_t r);
void instr32_0F49_reg(int32_t r1, int32_t r);
void instr16_0F4A_mem(int32_t addr, int32_t r);
void instr16_0F4A_reg(int32_t r1, int32_t r);
void instr32_0F4A_mem(int32_t addr, int32_t r);
void instr32_0F4A_reg(int32_t r1, int32_t r);
void instr16_0F4B_mem(int32_t addr, int32_t r);
void instr16_0F4B_reg(int32_t r1, int32_t r);
void instr32_0F4B_mem(int32_t addr, int32_t r);
void instr32_0F4B_reg(int32_t r1, int32_t r);
void instr16_0F4C_mem(int32_t addr, int32_t r);
void instr16_0F4C_reg(int32_t r1, int32_t r);
void instr32_0F4C_mem(int32_t addr, int32_t r);
void instr32_0F4C_reg(int32_t r1, int32_t r);
void instr16_0F4D_mem(int32_t addr, int32_t r);
void instr16_0F4D_reg(int32_t r1, int32_t r);
void instr32_0F4D_mem(int32_t addr, int32_t r);
void instr32_0F4D_reg(int32_t r1, int32_t r);
void instr16_0F4E_mem(int32_t addr, int32_t r);
void instr16_0F4E_reg(int32_t r1, int32_t r);
void instr32_0F4E_mem(int32_t addr, int32_t r);
void instr32_0F4E_reg(int32_t r1, int32_t r);
void instr16_0F4F_mem(int32_t addr, int32_t r);
void instr16_0F4F_reg(int32_t r1, int32_t r);
void instr32_0F4F_mem(int32_t addr, int32_t r);
void instr32_0F4F_reg(int32_t r1, int32_t r);
void instr_0F50_reg(int32_t r1, int32_t r2);
void instr_0F50_mem(int32_t addr, int32_t r1);
void instr_660F50_reg(int32_t r1, int32_t r2);
void instr_660F50_mem(int32_t addr, int32_t r1);
void instr_0F51(void);
void instr_0F52(void);
void instr_0F53(void);
void instr_0F54(union reg128 source, int32_t r);
void instr_0F54_reg(int32_t r1, int32_t r2);
void instr_0F54_mem(int32_t addr, int32_t r);
void instr_660F54(union reg128 source, int32_t r);
void instr_660F54_reg(int32_t r1, int32_t r2);
void instr_660F54_mem(int32_t addr, int32_t r);
void instr_0F55(union reg128 source, int32_t r);
void instr_0F55_reg(int32_t r1, int32_t r2);
void instr_0F55_mem(int32_t addr, int32_t r);
void instr_660F55(union reg128 source, int32_t r);
void instr_660F55_reg(int32_t r1, int32_t r2);
void instr_660F55_mem(int32_t addr, int32_t r);
void instr_0F56(union reg128 source, int32_t r);
void instr_0F56_reg(int32_t r1, int32_t r2);
void instr_0F56_mem(int32_t addr, int32_t r);
void instr_660F56(union reg128 source, int32_t r);
void instr_660F56_reg(int32_t r1, int32_t r2);
void instr_660F56_mem(int32_t addr, int32_t r);
void instr_0F57(union reg128 source, int32_t r);
void instr_0F57_reg(int32_t r1, int32_t r2);
void instr_0F57_mem(int32_t addr, int32_t r);
void instr_660F57(union reg128 source, int32_t r);
void instr_660F57_reg(int32_t r1, int32_t r2);
void instr_660F57_mem(int32_t addr, int32_t r);
void instr_0F58(void);
void instr_0F59(void);
void instr_0F5A(void);
void instr_0F5B(void);
void instr_0F5C(void);
void instr_0F5D(void);
void instr_0F5E(void);
void instr_0F5F(void);
void instr_0F60(int32_t source, int32_t r);
void instr_0F60_reg(int32_t r1, int32_t r2);
void instr_0F60_mem(int32_t addr, int32_t r);
void instr_660F60(union reg64 source, int32_t r);
void instr_660F60_reg(int32_t r1, int32_t r2);
void instr_660F60_mem(int32_t addr, int32_t r);
void instr_0F61(int32_t source, int32_t r);
void instr_0F61_reg(int32_t r1, int32_t r2);
void instr_0F61_mem(int32_t addr, int32_t r);
void instr_660F61(union reg64 source, int32_t r);
void instr_660F61_reg(int32_t r1, int32_t r2);
void instr_660F61_mem(int32_t addr, int32_t r);
void instr_0F62(int32_t source, int32_t r);
void instr_0F62_reg(int32_t r1, int32_t r2);
void instr_0F62_mem(int32_t addr, int32_t r);
void instr_660F62(union reg128 source, int32_t r);
void instr_660F62_reg(int32_t r1, int32_t r2);
void instr_660F62_mem(int32_t addr, int32_t r);
void instr_0F63(union reg64 source, int32_t r);
void instr_0F63_reg(int32_t r1, int32_t r2);
void instr_0F63_mem(int32_t addr, int32_t r);
void instr_660F63(union reg128 source, int32_t r);
void instr_660F63_reg(int32_t r1, int32_t r2);
void instr_660F63_mem(int32_t addr, int32_t r);
void instr_0F64(union reg64 source, int32_t r);
void instr_0F64_reg(int32_t r1, int32_t r2);
void instr_0F64_mem(int32_t addr, int32_t r);
void instr_660F64(union reg128 source, int32_t r);
void instr_660F64_reg(int32_t r1, int32_t r2);
void instr_660F64_mem(int32_t addr, int32_t r);
void instr_0F65(union reg64 source, int32_t r);
void instr_0F65_reg(int32_t r1, int32_t r2);
void instr_0F65_mem(int32_t addr, int32_t r);
void instr_660F65(union reg128 source, int32_t r);
void instr_660F65_reg(int32_t r1, int32_t r2);
void instr_660F65_mem(int32_t addr, int32_t r);
void instr_0F66(union reg64 source, int32_t r);
void instr_0F66_reg(int32_t r1, int32_t r2);
void instr_0F66_mem(int32_t addr, int32_t r);
void instr_660F66(union reg128 source, int32_t r);
void instr_660F66_reg(int32_t r1, int32_t r2);
void instr_660F66_mem(int32_t addr, int32_t r);
void instr_0F67(union reg64 source, int32_t r);
void instr_0F67_reg(int32_t r1, int32_t r2);
void instr_0F67_mem(int32_t addr, int32_t r);
void instr_660F67(union reg128 source, int32_t r);
void instr_660F67_reg(int32_t r1, int32_t r2);
void instr_660F67_mem(int32_t addr, int32_t r);
void instr_0F68(union reg64 source, int32_t r);
void instr_0F68_reg(int32_t r1, int32_t r2);
void instr_0F68_mem(int32_t addr, int32_t r);
void instr_660F68(union reg128 source, int32_t r);
void instr_660F68_reg(int32_t r1, int32_t r2);
void instr_660F68_mem(int32_t addr, int32_t r);
void instr_0F69(union reg64 source, int32_t r);
void instr_0F69_reg(int32_t r1, int32_t r2);
void instr_0F69_mem(int32_t addr, int32_t r);
void instr_660F69(union reg128 source, int32_t r);
void instr_660F69_reg(int32_t r1, int32_t r2);
void instr_660F69_mem(int32_t addr, int32_t r);
void instr_0F6A(union reg64 source, int32_t r);
void instr_0F6A_reg(int32_t r1, int32_t r2);
void instr_0F6A_mem(int32_t addr, int32_t r);
void instr_660F6A(union reg128 source, int32_t r);
void instr_660F6A_reg(int32_t r1, int32_t r2);
void instr_660F6A_mem(int32_t addr, int32_t r);
void instr_0F6B(union reg64 source, int32_t r);
void instr_0F6B_reg(int32_t r1, int32_t r2);
void instr_0F6B_mem(int32_t addr, int32_t r);
void instr_660F6B(union reg128 source, int32_t r);
void instr_660F6B_reg(int32_t r1, int32_t r2);
void instr_660F6B_mem(int32_t addr, int32_t r);
void instr_0F6C_mem(int32_t addr, int32_t r);
void instr_0F6C_reg(int32_t r1, int32_t r2);
void instr_660F6C(union reg128 source, int32_t r);
void instr_660F6C_reg(int32_t r1, int32_t r2);
void instr_660F6C_mem(int32_t addr, int32_t r);
void instr_0F6D_mem(int32_t addr, int32_t r);
void instr_0F6D_reg(int32_t r1, int32_t r2);
void instr_660F6D(union reg128 source, int32_t r);
void instr_660F6D_reg(int32_t r1, int32_t r2);
void instr_660F6D_mem(int32_t addr, int32_t r);
void instr_0F6E(int32_t source, int32_t r);
void instr_0F6E_reg(int32_t r1, int32_t r2);
void instr_0F6E_mem(int32_t addr, int32_t r);
void instr_660F6E(int32_t source, int32_t r);
void instr_660F6E_reg(int32_t r1, int32_t r2);
void instr_660F6E_mem(int32_t addr, int32_t r);
void instr_0F6F(union reg64 source, int32_t r);
void instr_0F6F_reg(int32_t r1, int32_t r2);
void instr_0F6F_mem(int32_t addr, int32_t r);
void instr_660F6F(union reg128 source, int32_t r);
void instr_660F6F_reg(int32_t r1, int32_t r2);
void instr_660F6F_mem(int32_t addr, int32_t r);
void instr_F30F6F(union reg128 source, int32_t r);
void instr_F30F6F_reg(int32_t r1, int32_t r2);
void instr_F30F6F_mem(int32_t addr, int32_t r);
void instr_0F70(union reg64 source, int32_t r, int32_t imm8);
void instr_0F70_reg(int32_t r1, int32_t r2, int32_t imm);
void instr_0F70_mem(int32_t addr, int32_t r, int32_t imm);
void instr_660F70(union reg128 source, int32_t r, int32_t imm8);
void instr_660F70_reg(int32_t r1, int32_t r2, int32_t imm);
void instr_660F70_mem(int32_t addr, int32_t r, int32_t imm);
void instr_F20F70(union reg128 source, int32_t r, int32_t imm8);
void instr_F20F70_reg(int32_t r1, int32_t r2, int32_t imm);
void instr_F20F70_mem(int32_t addr, int32_t r, int32_t imm);
void instr_F30F70(union reg128 source, int32_t r, int32_t imm8);
void instr_F30F70_reg(int32_t r1, int32_t r2, int32_t imm);
void instr_F30F70_mem(int32_t addr, int32_t r, int32_t imm);
void instr_0F71_2_mem(int32_t addr, int32_t r);
void instr_0F71_4_mem(int32_t addr, int32_t r);
void instr_0F71_6_mem(int32_t addr, int32_t r);
void instr_0F71_2_reg(int32_t r, int32_t imm8);
void instr_0F71_4_reg(int32_t r, int32_t imm8);
void instr_0F71_6_reg(int32_t r, int32_t imm8);
void instr_660F71_2_mem(int32_t addr, int32_t r);
void instr_660F71_4_mem(int32_t addr, int32_t r);
void instr_660F71_6_mem(int32_t addr, int32_t r);
void instr_660F71_2_reg(int32_t r, int32_t imm8);
void instr_660F71_4_reg(int32_t r, int32_t imm8);
void instr_660F71_6_reg(int32_t r, int32_t imm8);
void instr_0F72_2_mem(int32_t addr, int32_t r);
void instr_0F72_4_mem(int32_t addr, int32_t r);
void instr_0F72_6_mem(int32_t addr, int32_t r);
void instr_0F72_2_reg(int32_t r, int32_t imm8);
void instr_0F72_4_reg(int32_t r, int32_t imm8);
void instr_0F72_6_reg(int32_t r, int32_t imm8);
void instr_660F72_2_mem(int32_t addr, int32_t r);
void instr_660F72_4_mem(int32_t addr, int32_t r);
void instr_660F72_6_mem(int32_t addr, int32_t r);
void instr_660F72_2_reg(int32_t r, int32_t imm8);
void instr_660F72_4_reg(int32_t r, int32_t imm8);
void instr_660F72_6_reg(int32_t r, int32_t imm8);
void instr_0F73_2_mem(int32_t addr, int32_t r);
void instr_0F73_3_mem(int32_t addr, int32_t r);
void instr_0F73_3_reg(int32_t addr, int32_t r);
void instr_0F73_6_mem(int32_t addr, int32_t r);
void instr_0F73_7_mem(int32_t addr, int32_t r);
void instr_0F73_7_reg(int32_t addr, int32_t r);
void instr_0F73_2_reg(int32_t r, int32_t imm8);
void instr_0F73_6_reg(int32_t r, int32_t imm8);
void instr_660F73_2_mem(int32_t addr, int32_t r);
void instr_660F73_3_mem(int32_t addr, int32_t r);
void instr_660F73_6_mem(int32_t addr, int32_t r);
void instr_660F73_7_mem(int32_t addr, int32_t r);
void instr_660F73_2_reg(int32_t r, int32_t imm8);
void instr_660F73_3_reg(int32_t r, int32_t imm8);
void instr_660F73_6_reg(int32_t r, int32_t imm8);
void instr_660F73_7_reg(int32_t r, int32_t imm8);
void instr_0F74(union reg64 source, int32_t r);
void instr_0F74_reg(int32_t r1, int32_t r2);
void instr_0F74_mem(int32_t addr, int32_t r);
void instr_660F74(union reg128 source, int32_t r);
void instr_660F74_reg(int32_t r1, int32_t r2);
void instr_660F74_mem(int32_t addr, int32_t r);
void instr_0F75(union reg64 source, int32_t r);
void instr_0F75_reg(int32_t r1, int32_t r2);
void instr_0F75_mem(int32_t addr, int32_t r);
void instr_660F75(union reg128 source, int32_t r);
void instr_660F75_reg(int32_t r1, int32_t r2);
void instr_660F75_mem(int32_t addr, int32_t r);
void instr_0F76(union reg64 source, int32_t r);
void instr_0F76_reg(int32_t r1, int32_t r2);
void instr_0F76_mem(int32_t addr, int32_t r);
void instr_660F76(union reg128 source, int32_t r);
void instr_660F76_reg(int32_t r1, int32_t r2);
void instr_660F76_mem(int32_t addr, int32_t r);
void instr_0F77(void);
void instr_0F78(void);
void instr_0F79(void);
void instr_0F7A(void);
void instr_0F7B(void);
void instr_0F7C(void);
void instr_0F7D(void);
int32_t instr_0F7E(int32_t r);
void instr_0F7E_reg(int32_t r1, int32_t r2);
void instr_0F7E_mem(int32_t addr, int32_t r);
int32_t instr_660F7E(int32_t r);
void instr_660F7E_reg(int32_t r1, int32_t r2);
void instr_660F7E_mem(int32_t addr, int32_t r);
void instr_F30F7E_mem(int32_t addr, int32_t r);
void instr_F30F7E_reg(int32_t r1, int32_t r2);
void instr_0F7F_mem(int32_t addr, int32_t r);
void instr_0F7F_reg(int32_t r1, int32_t r2);
void instr_660F7F_mem(int32_t addr, int32_t r);
void instr_660F7F_reg(int32_t r1, int32_t r2);
void instr_F30F7F_mem(int32_t addr, int32_t r);
void instr_F30F7F_reg(int32_t r1, int32_t r2);
void instr16_0F80(int32_t imm);
void instr32_0F80(int32_t imm);
void instr16_0F81(int32_t imm);
void instr32_0F81(int32_t imm);
void instr16_0F82(int32_t imm);
void instr32_0F82(int32_t imm);
void instr16_0F83(int32_t imm);
void instr32_0F83(int32_t imm);
void instr16_0F84(int32_t imm);
void instr32_0F84(int32_t imm);
void instr16_0F85(int32_t imm);
void instr32_0F85(int32_t imm);
void instr16_0F86(int32_t imm);
void instr32_0F86(int32_t imm);
void instr16_0F87(int32_t imm);
void instr32_0F87(int32_t imm);
void instr16_0F88(int32_t imm);
void instr32_0F88(int32_t imm);
void instr16_0F89(int32_t imm);
void instr32_0F89(int32_t imm);
void instr16_0F8A(int32_t imm);
void instr32_0F8A(int32_t imm);
void instr16_0F8B(int32_t imm);
void instr32_0F8B(int32_t imm);
void instr16_0F8C(int32_t imm);
void instr32_0F8C(int32_t imm);
void instr16_0F8D(int32_t imm);
void instr32_0F8D(int32_t imm);
void instr16_0F8E(int32_t imm);
void instr32_0F8E(int32_t imm);
void instr16_0F8F(int32_t imm);
void instr32_0F8F(int32_t imm);
void instr_0F90_reg(int32_t r, int32_t unused);
void instr_0F91_reg(int32_t r, int32_t unused);
void instr_0F92_reg(int32_t r, int32_t unused);
void instr_0F93_reg(int32_t r, int32_t unused);
void instr_0F94_reg(int32_t r, int32_t unused);
void instr_0F95_reg(int32_t r, int32_t unused);
void instr_0F96_reg(int32_t r, int32_t unused);
void instr_0F97_reg(int32_t r, int32_t unused);
void instr_0F98_reg(int32_t r, int32_t unused);
void instr_0F99_reg(int32_t r, int32_t unused);
void instr_0F9A_reg(int32_t r, int32_t unused);
void instr_0F9B_reg(int32_t r, int32_t unused);
void instr_0F9C_reg(int32_t r, int32_t unused);
void instr_0F9D_reg(int32_t r, int32_t unused);
void instr_0F9E_reg(int32_t r, int32_t unused);
void instr_0F9F_reg(int32_t r, int32_t unused);
void instr_0F90_mem(int32_t addr, int32_t unused);
void instr_0F91_mem(int32_t addr, int32_t unused);
void instr_0F92_mem(int32_t addr, int32_t unused);
void instr_0F93_mem(int32_t addr, int32_t unused);
void instr_0F94_mem(int32_t addr, int32_t unused);
void instr_0F95_mem(int32_t addr, int32_t unused);
void instr_0F96_mem(int32_t addr, int32_t unused);
void instr_0F97_mem(int32_t addr, int32_t unused);
void instr_0F98_mem(int32_t addr, int32_t unused);
void instr_0F99_mem(int32_t addr, int32_t unused);
void instr_0F9A_mem(int32_t addr, int32_t unused);
void instr_0F9B_mem(int32_t addr, int32_t unused);
void instr_0F9C_mem(int32_t addr, int32_t unused);
void instr_0F9D_mem(int32_t addr, int32_t unused);
void instr_0F9E_mem(int32_t addr, int32_t unused);
void instr_0F9F_mem(int32_t addr, int32_t unused);
void instr16_0FA0(void);
void instr32_0FA0(void);
void instr16_0FA1(void);
void instr32_0FA1(void);
void instr_0FA2(void);
void instr16_0FA3_reg(int32_t r1, int32_t r2);
void instr16_0FA3_mem(int32_t addr, int32_t r);
void instr32_0FA3_reg(int32_t r1, int32_t r2);
void instr32_0FA3_mem(int32_t addr, int32_t r);
void instr16_0FA4_mem(int32_t addr, int32_t r, int32_t imm);
void instr16_0FA4_reg(int32_t r1, int32_t r, int32_t imm);
void instr32_0FA4_mem(int32_t addr, int32_t r, int32_t imm);
void instr32_0FA4_reg(int32_t r1, int32_t r, int32_t imm);
void instr16_0FA5_mem(int32_t addr, int32_t r);
void instr16_0FA5_reg(int32_t r1, int32_t r);
void instr32_0FA5_mem(int32_t addr, int32_t r);
void instr32_0FA5_reg(int32_t r1, int32_t r);
void instr_0FA6(void);
void instr_0FA7(void);
void instr16_0FA8(void);
void instr32_0FA8(void);
void instr16_0FA9(void);
void instr32_0FA9(void);
void instr_0FAA(void);
void instr16_0FAB_reg(int32_t r1, int32_t r2);
void instr16_0FAB_mem(int32_t addr, int32_t r);
void instr32_0FAB_reg(int32_t r1, int32_t r2);
void instr32_0FAB_mem(int32_t addr, int32_t r);
void instr16_0FAC_mem(int32_t addr, int32_t r, int32_t imm);
void instr16_0FAC_reg(int32_t r1, int32_t r, int32_t imm);
void instr32_0FAC_mem(int32_t addr, int32_t r, int32_t imm);
void instr32_0FAC_reg(int32_t r1, int32_t r, int32_t imm);
void instr16_0FAD_mem(int32_t addr, int32_t r);
void instr16_0FAD_reg(int32_t r1, int32_t r);
void instr32_0FAD_mem(int32_t addr, int32_t r);
void instr32_0FAD_reg(int32_t r1, int32_t r);
void instr_0FAE_0_reg(int32_t r);
void instr_0FAE_0_mem(int32_t addr);
void instr_0FAE_1_reg(int32_t r);
void instr_0FAE_1_mem(int32_t addr);
void instr_0FAE_2_reg(int32_t r);
void instr_0FAE_2_mem(int32_t addr);
void instr_0FAE_3_reg(int32_t r);
void instr_0FAE_3_mem(int32_t addr);
void instr_0FAE_4_reg(int32_t r);
void instr_0FAE_4_mem(int32_t addr);
void instr_0FAE_5_reg(int32_t r);
void instr_0FAE_5_mem(int32_t addr);
void instr_0FAE_6_reg(int32_t r);
void instr_0FAE_6_mem(int32_t addr);
void instr_0FAE_7_reg(int32_t r);
void instr_0FAE_7_mem(int32_t addr);
void instr16_0FAF_mem(int32_t addr, int32_t r);
void instr16_0FAF_reg(int32_t r1, int32_t r);
void instr32_0FAF_mem(int32_t addr, int32_t r);
void instr32_0FAF_reg(int32_t r1, int32_t r);
void instr_0FB0_reg(int32_t r1, int32_t r2);
void instr_0FB0_mem(int32_t addr, int32_t r);
void instr16_0FB1_reg(int32_t r1, int32_t r2);
void instr16_0FB1_mem(int32_t addr, int32_t r);
void instr32_0FB1_reg(int32_t r1, int32_t r2);
void instr32_0FB1_mem(int32_t addr, int32_t r);
void instr16_0FB2_reg(int32_t unused, int32_t unused2);
void instr16_0FB2_mem(int32_t addr, int32_t r);
void instr32_0FB2_reg(int32_t unused, int32_t unused2);
void instr32_0FB2_mem(int32_t addr, int32_t r);
void instr16_0FB3_reg(int32_t r1, int32_t r2);
void instr16_0FB3_mem(int32_t addr, int32_t r);
void instr32_0FB3_reg(int32_t r1, int32_t r2);
void instr32_0FB3_mem(int32_t addr, int32_t r);
void instr16_0FB4_reg(int32_t unused, int32_t unused2);
void instr16_0FB4_mem(int32_t addr, int32_t r);
void instr32_0FB4_reg(int32_t unused, int32_t unused2);
void instr32_0FB4_mem(int32_t addr, int32_t r);
void instr16_0FB5_reg(int32_t unused, int32_t unused2);
void instr16_0FB5_mem(int32_t addr, int32_t r);
void instr32_0FB5_reg(int32_t unused, int32_t unused2);
void instr32_0FB5_mem(int32_t addr, int32_t r);
void instr16_0FB6_mem(int32_t addr, int32_t r);
void instr16_0FB6_reg(int32_t r1, int32_t r);
void instr32_0FB6_mem(int32_t addr, int32_t r);
void instr32_0FB6_reg(int32_t r1, int32_t r);
void instr16_0FB7_mem(int32_t addr, int32_t r);
void instr16_0FB7_reg(int32_t r1, int32_t r);
void instr32_0FB7_mem(int32_t addr, int32_t r);
void instr32_0FB7_reg(int32_t r1, int32_t r);
void instr16_0FB8_reg(int32_t r1, int32_t r2);
void instr16_0FB8_mem(int32_t addr, int32_t r);
void instr16_F30FB8_mem(int32_t addr, int32_t r);
void instr16_F30FB8_reg(int32_t r1, int32_t r);
void instr32_0FB8_reg(int32_t r1, int32_t r2);
void instr32_0FB8_mem(int32_t addr, int32_t r);
void instr32_F30FB8_mem(int32_t addr, int32_t r);
void instr32_F30FB8_reg(int32_t r1, int32_t r);
void instr_0FB9(void);
void instr16_0FBA_4_reg(int32_t r, int32_t imm);
void instr16_0FBA_4_mem(int32_t addr, int32_t imm);
void instr16_0FBA_5_reg(int32_t r, int32_t imm);
void instr16_0FBA_5_mem(int32_t addr, int32_t imm);
void instr16_0FBA_6_reg(int32_t r, int32_t imm);
void instr16_0FBA_6_mem(int32_t addr, int32_t imm);
void instr16_0FBA_7_reg(int32_t r, int32_t imm);
void instr16_0FBA_7_mem(int32_t addr, int32_t imm);
void instr32_0FBA_4_reg(int32_t r, int32_t imm);
void instr32_0FBA_4_mem(int32_t addr, int32_t imm);
void instr32_0FBA_5_reg(int32_t r, int32_t imm);
void instr32_0FBA_5_mem(int32_t addr, int32_t imm);
void instr32_0FBA_6_reg(int32_t r, int32_t imm);
void instr32_0FBA_6_mem(int32_t addr, int32_t imm);
void instr32_0FBA_7_reg(int32_t r, int32_t imm);
void instr32_0FBA_7_mem(int32_t addr, int32_t imm);
void instr16_0FBB_reg(int32_t r1, int32_t r2);
void instr16_0FBB_mem(int32_t addr, int32_t r);
void instr32_0FBB_reg(int32_t r1, int32_t r2);
void instr32_0FBB_mem(int32_t addr, int32_t r);
void instr16_0FBC_mem(int32_t addr, int32_t r);
void instr16_0FBC_reg(int32_t r1, int32_t r);
void instr32_0FBC_mem(int32_t addr, int32_t r);
void instr32_0FBC_reg(int32_t r1, int32_t r);
void instr16_0FBD_mem(int32_t addr, int32_t r);
void instr16_0FBD_reg(int32_t r1, int32_t r);
void instr32_0FBD_mem(int32_t addr, int32_t r);
void instr32_0FBD_reg(int32_t r1, int32_t r);
void instr16_0FBE_mem(int32_t addr, int32_t r);
void instr16_0FBE_reg(int32_t r1, int32_t r);
void instr32_0FBE_mem(int32_t addr, int32_t r);
void instr32_0FBE_reg(int32_t r1, int32_t r);
void instr16_0FBF_mem(int32_t addr, int32_t r);
void instr16_0FBF_reg(int32_t r1, int32_t r);
void instr32_0FBF_mem(int32_t addr, int32_t r);
void instr32_0FBF_reg(int32_t r1, int32_t r);
void instr_0FC0_mem(int32_t addr, int32_t r);
void instr_0FC0_reg(int32_t r1, int32_t r);
void instr16_0FC1_mem(int32_t addr, int32_t r);
void instr16_0FC1_reg(int32_t r1, int32_t r);
void instr32_0FC1_mem(int32_t addr, int32_t r);
void instr32_0FC1_reg(int32_t r1, int32_t r);
void instr_0FC2(void);
void instr_0FC3_reg(int32_t r1, int32_t r2);
void instr_0FC3_mem(int32_t addr, int32_t r);
void instr_0FC4(int32_t source, int32_t r, int32_t imm8);
void instr_0FC4_reg(int32_t r1, int32_t r2, int32_t imm);
void instr_0FC4_mem(int32_t addr, int32_t r, int32_t imm);
void instr_660FC4(int32_t source, int32_t r, int32_t imm8);
void instr_660FC4_reg(int32_t r1, int32_t r2, int32_t imm);
void instr_660FC4_mem(int32_t addr, int32_t r, int32_t imm);
void instr_0FC5_mem(int32_t addr, int32_t r, int32_t imm8);
void instr_0FC5_reg(int32_t r1, int32_t r2, int32_t imm8);
void instr_660FC5_mem(int32_t addr, int32_t r, int32_t imm8);
void instr_660FC5_reg(int32_t r1, int32_t r2, int32_t imm8);
void instr_0FC6(void);
void instr_0FC7_1_reg(int32_t r);
void instr_0FC7_1_mem(int32_t addr);
void instr_0FC7_6_reg(int32_t r);
void instr_0FC7_6_mem(int32_t addr);
void instr_0FC8(void);
void instr_0FC9(void);
void instr_0FCA(void);
void instr_0FCB(void);
void instr_0FCC(void);
void instr_0FCD(void);
void instr_0FCE(void);
void instr_0FCF(void);
void instr_0FD0(void);
void instr_0FD1(union reg64 source, int32_t r);
void instr_0FD1_reg(int32_t r1, int32_t r2);
void instr_0FD1_mem(int32_t addr, int32_t r);
void instr_660FD1(union reg128 source, int32_t r);
void instr_660FD1_reg(int32_t r1, int32_t r2);
void instr_660FD1_mem(int32_t addr, int32_t r);
void instr_0FD2(union reg64 source, int32_t r);
void instr_0FD2_reg(int32_t r1, int32_t r2);
void instr_0FD2_mem(int32_t addr, int32_t r);
void instr_660FD2(union reg128 source, int32_t r);
void instr_660FD2_reg(int32_t r1, int32_t r2);
void instr_660FD2_mem(int32_t addr, int32_t r);
void instr_0FD3(union reg64 source, int32_t r);
void instr_0FD3_reg(int32_t r1, int32_t r2);
void instr_0FD3_mem(int32_t addr, int32_t r);
void instr_660FD3(union reg128 source, int32_t r);
void instr_660FD3_reg(int32_t r1, int32_t r2);
void instr_660FD3_mem(int32_t addr, int32_t r);
void instr_0FD4(union reg64 source, int32_t r);
void instr_0FD4_reg(int32_t r1, int32_t r2);
void instr_0FD4_mem(int32_t addr, int32_t r);
void instr_660FD4(union reg128 source, int32_t r);
void instr_660FD4_reg(int32_t r1, int32_t r2);
void instr_660FD4_mem(int32_t addr, int32_t r);
void instr_0FD5(union reg64 source, int32_t r);
void instr_0FD5_reg(int32_t r1, int32_t r2);
void instr_0FD5_mem(int32_t addr, int32_t r);
void instr_660FD5(union reg128 source, int32_t r);
void instr_660FD5_reg(int32_t r1, int32_t r2);
void instr_660FD5_mem(int32_t addr, int32_t r);
void instr_0FD6_mem(int32_t addr, int32_t r);
void instr_0FD6_reg(int32_t r1, int32_t r2);
void instr_660FD6_mem(int32_t addr, int32_t r);
void instr_660FD6_reg(int32_t r1, int32_t r2);
void instr_F20FD6_mem(int32_t addr, int32_t r);
void instr_F20FD6_reg(int32_t r1, int32_t r2);
void instr_F30FD6_mem(int32_t addr, int32_t r);
void instr_F30FD6_reg(int32_t r1, int32_t r2);
void instr_0FD7_mem(int32_t addr, int32_t r);
void instr_0FD7_reg(int32_t r1, int32_t r2);
void instr_660FD7_mem(int32_t addr, int32_t r);
void instr_660FD7_reg(int32_t r1, int32_t r2);
void instr_0FD8(union reg64 source, int32_t r);
void instr_0FD8_reg(int32_t r1, int32_t r2);
void instr_0FD8_mem(int32_t addr, int32_t r);
void instr_660FD8(union reg128 source, int32_t r);
void instr_660FD8_reg(int32_t r1, int32_t r2);
void instr_660FD8_mem(int32_t addr, int32_t r);
void instr_0FD9(union reg64 source, int32_t r);
void instr_0FD9_reg(int32_t r1, int32_t r2);
void instr_0FD9_mem(int32_t addr, int32_t r);
void instr_660FD9(union reg128 source, int32_t r);
void instr_660FD9_reg(int32_t r1, int32_t r2);
void instr_660FD9_mem(int32_t addr, int32_t r);
void instr_0FDA(union reg64 source, int32_t r);
void instr_0FDA_reg(int32_t r1, int32_t r2);
void instr_0FDA_mem(int32_t addr, int32_t r);
void instr_660FDA(union reg128 source, int32_t r);
void instr_660FDA_reg(int32_t r1, int32_t r2);
void instr_660FDA_mem(int32_t addr, int32_t r);
void instr_0FDB(union reg64 source, int32_t r);
void instr_0FDB_reg(int32_t r1, int32_t r2);
void instr_0FDB_mem(int32_t addr, int32_t r);
void instr_660FDB(union reg128 source, int32_t r);
void instr_660FDB_reg(int32_t r1, int32_t r2);
void instr_660FDB_mem(int32_t addr, int32_t r);
void instr_0FDC(union reg64 source, int32_t r);
void instr_0FDC_reg(int32_t r1, int32_t r2);
void instr_0FDC_mem(int32_t addr, int32_t r);
void instr_660FDC(union reg128 source, int32_t r);
void instr_660FDC_reg(int32_t r1, int32_t r2);
void instr_660FDC_mem(int32_t addr, int32_t r);
void instr_0FDD(union reg64 source, int32_t r);
void instr_0FDD_reg(int32_t r1, int32_t r2);
void instr_0FDD_mem(int32_t addr, int32_t r);
void instr_660FDD(union reg128 source, int32_t r);
void instr_660FDD_reg(int32_t r1, int32_t r2);
void instr_660FDD_mem(int32_t addr, int32_t r);
void instr_0FDE(union reg64 source, int32_t r);
void instr_0FDE_reg(int32_t r1, int32_t r2);
void instr_0FDE_mem(int32_t addr, int32_t r);
void instr_660FDE(union reg128 source, int32_t r);
void instr_660FDE_reg(int32_t r1, int32_t r2);
void instr_660FDE_mem(int32_t addr, int32_t r);
void instr_0FDF(union reg64 source, int32_t r);
void instr_0FDF_reg(int32_t r1, int32_t r2);
void instr_0FDF_mem(int32_t addr, int32_t r);
void instr_660FDF(union reg128 source, int32_t r);
void instr_660FDF_reg(int32_t r1, int32_t r2);
void instr_660FDF_mem(int32_t addr, int32_t r);
void instr_0FE0(union reg64 source, int32_t r);
void instr_0FE0_reg(int32_t r1, int32_t r2);
void instr_0FE0_mem(int32_t addr, int32_t r);
void instr_660FE0(union reg128 source, int32_t r);
void instr_660FE0_reg(int32_t r1, int32_t r2);
void instr_660FE0_mem(int32_t addr, int32_t r);
void instr_0FE1(union reg64 source, int32_t r);
void instr_0FE1_reg(int32_t r1, int32_t r2);
void instr_0FE1_mem(int32_t addr, int32_t r);
void instr_660FE1(union reg128 source, int32_t r);
void instr_660FE1_reg(int32_t r1, int32_t r2);
void instr_660FE1_mem(int32_t addr, int32_t r);
void instr_0FE2(union reg64 source, int32_t r);
void instr_0FE2_reg(int32_t r1, int32_t r2);
void instr_0FE2_mem(int32_t addr, int32_t r);
void instr_660FE2(union reg128 source, int32_t r);
void instr_660FE2_reg(int32_t r1, int32_t r2);
void instr_660FE2_mem(int32_t addr, int32_t r);
void instr_0FE3(union reg64 source, int32_t r);
void instr_0FE3_reg(int32_t r1, int32_t r2);
void instr_0FE3_mem(int32_t addr, int32_t r);
void instr_660FE3(union reg128 source, int32_t r);
void instr_660FE3_reg(int32_t r1, int32_t r2);
void instr_660FE3_mem(int32_t addr, int32_t r);
void instr_0FE4(union reg64 source, int32_t r);
void instr_0FE4_reg(int32_t r1, int32_t r2);
void instr_0FE4_mem(int32_t addr, int32_t r);
void instr_660FE4(union reg128 source, int32_t r);
void instr_660FE4_reg(int32_t r1, int32_t r2);
void instr_660FE4_mem(int32_t addr, int32_t r);
void instr_0FE5(union reg64 source, int32_t r);
void instr_0FE5_reg(int32_t r1, int32_t r2);
void instr_0FE5_mem(int32_t addr, int32_t r);
void instr_660FE5(union reg128 source, int32_t r);
void instr_660FE5_reg(int32_t r1, int32_t r2);
void instr_660FE5_mem(int32_t addr, int32_t r);
void instr_0FE6_mem(int32_t addr, int32_t r);
void instr_0FE6_reg(int32_t r1, int32_t r2);
void instr_660FE6_mem(int32_t addr, int32_t r);
void instr_660FE6_reg(int32_t r1, int32_t r2);
void instr_F20FE6_mem(int32_t addr, int32_t r);
void instr_F20FE6_reg(int32_t r1, int32_t r2);
void instr_F30FE6_mem(int32_t addr, int32_t r);
void instr_F30FE6_reg(int32_t r1, int32_t r2);
void instr_0FE7_mem(int32_t addr, int32_t r);
void instr_0FE7_reg(int32_t r1, int32_t r2);
void instr_660FE7_reg(int32_t r1, int32_t r2);
void instr_660FE7_mem(int32_t addr, int32_t r);
void instr_0FE8(union reg64 source, int32_t r);
void instr_0FE8_reg(int32_t r1, int32_t r2);
void instr_0FE8_mem(int32_t addr, int32_t r);
void instr_660FE8(union reg128 source, int32_t r);
void instr_660FE8_reg(int32_t r1, int32_t r2);
void instr_660FE8_mem(int32_t addr, int32_t r);
void instr_0FE9(union reg64 source, int32_t r);
void instr_0FE9_reg(int32_t r1, int32_t r2);
void instr_0FE9_mem(int32_t addr, int32_t r);
void instr_660FE9(union reg128 source, int32_t r);
void instr_660FE9_reg(int32_t r1, int32_t r2);
void instr_660FE9_mem(int32_t addr, int32_t r);
void instr_0FEA(union reg64 source, int32_t r);
void instr_0FEA_reg(int32_t r1, int32_t r2);
void instr_0FEA_mem(int32_t addr, int32_t r);
void instr_660FEA(union reg128 source, int32_t r);
void instr_660FEA_reg(int32_t r1, int32_t r2);
void instr_660FEA_mem(int32_t addr, int32_t r);
void instr_0FEB(union reg64 source, int32_t r);
void instr_0FEB_reg(int32_t r1, int32_t r2);
void instr_0FEB_mem(int32_t addr, int32_t r);
void instr_660FEB(union reg128 source, int32_t r);
void instr_660FEB_reg(int32_t r1, int32_t r2);
void instr_660FEB_mem(int32_t addr, int32_t r);
void instr_0FEC(union reg64 source, int32_t r);
void instr_0FEC_reg(int32_t r1, int32_t r2);
void instr_0FEC_mem(int32_t addr, int32_t r);
void instr_660FEC(union reg128 source, int32_t r);
void instr_660FEC_reg(int32_t r1, int32_t r2);
void instr_660FEC_mem(int32_t addr, int32_t r);
void instr_0FED(union reg64 source, int32_t r);
void instr_0FED_reg(int32_t r1, int32_t r2);
void instr_0FED_mem(int32_t addr, int32_t r);
void instr_660FED(union reg128 source, int32_t r);
void instr_660FED_reg(int32_t r1, int32_t r2);
void instr_660FED_mem(int32_t addr, int32_t r);
void instr_0FEE(union reg64 source, int32_t r);
void instr_0FEE_reg(int32_t r1, int32_t r2);
void instr_0FEE_mem(int32_t addr, int32_t r);
void instr_660FEE(union reg128 source, int32_t r);
void instr_660FEE_reg(int32_t r1, int32_t r2);
void instr_660FEE_mem(int32_t addr, int32_t r);
void instr_0FEF(union reg64 source, int32_t r);
void instr_0FEF_reg(int32_t r1, int32_t r2);
void instr_0FEF_mem(int32_t addr, int32_t r);
void instr_660FEF(union reg128 source, int32_t r);
void instr_660FEF_reg(int32_t r1, int32_t r2);
void instr_660FEF_mem(int32_t addr, int32_t r);
void instr_0FF0(void);
void instr_0FF1(union reg64 source, int32_t r);
void instr_0FF1_reg(int32_t r1, int32_t r2);
void instr_0FF1_mem(int32_t addr, int32_t r);
void instr_660FF1(union reg128 source, int32_t r);
void instr_660FF1_reg(int32_t r1, int32_t r2);
void instr_660FF1_mem(int32_t addr, int32_t r);
void instr_0FF2(union reg64 source, int32_t r);
void instr_0FF2_reg(int32_t r1, int32_t r2);
void instr_0FF2_mem(int32_t addr, int32_t r);
void instr_660FF2(union reg128 source, int32_t r);
void instr_660FF2_reg(int32_t r1, int32_t r2);
void instr_660FF2_mem(int32_t addr, int32_t r);
void instr_0FF3(union reg64 source, int32_t r);
void instr_0FF3_reg(int32_t r1, int32_t r2);
void instr_0FF3_mem(int32_t addr, int32_t r);
void instr_660FF3(union reg128 source, int32_t r);
void instr_660FF3_reg(int32_t r1, int32_t r2);
void instr_660FF3_mem(int32_t addr, int32_t r);
void instr_0FF4(union reg64 source, int32_t r);
void instr_0FF4_reg(int32_t r1, int32_t r2);
void instr_0FF4_mem(int32_t addr, int32_t r);
void instr_660FF4(union reg128 source, int32_t r);
void instr_660FF4_reg(int32_t r1, int32_t r2);
void instr_660FF4_mem(int32_t addr, int32_t r);
void instr_0FF5(union reg64 source, int32_t r);
void instr_0FF5_reg(int32_t r1, int32_t r2);
void instr_0FF5_mem(int32_t addr, int32_t r);
void instr_660FF5(union reg128 source, int32_t r);
void instr_660FF5_reg(int32_t r1, int32_t r2);
void instr_660FF5_mem(int32_t addr, int32_t r);
void instr_0FF6(union reg64 source, int32_t r);
void instr_0FF6_reg(int32_t r1, int32_t r2);
void instr_0FF6_mem(int32_t addr, int32_t r);
void instr_660FF6(union reg128 source, int32_t r);
void instr_660FF6_reg(int32_t r1, int32_t r2);
void instr_660FF6_mem(int32_t addr, int32_t r);
void instr_0FF7_mem(int32_t addr, int32_t r);
void instr_0FF7_reg(int32_t r1, int32_t r2);
void instr_660FF7_mem(int32_t addr, int32_t r);
void instr_660FF7_reg(int32_t r1, int32_t r2);
void instr_0FF8(union reg64 source, int32_t r);
void instr_0FF8_reg(int32_t r1, int32_t r2);
void instr_0FF8_mem(int32_t addr, int32_t r);
void instr_660FF8(union reg128 source, int32_t r);
void instr_660FF8_reg(int32_t r1, int32_t r2);
void instr_660FF8_mem(int32_t addr, int32_t r);
void instr_0FF9(union reg64 source, int32_t r);
void instr_0FF9_reg(int32_t r1, int32_t r2);
void instr_0FF9_mem(int32_t addr, int32_t r);
void instr_660FF9(union reg128 source, int32_t r);
void instr_660FF9_reg(int32_t r1, int32_t r2);
void instr_660FF9_mem(int32_t addr, int32_t r);
void instr_0FFA(union reg64 source, int32_t r);
void instr_0FFA_reg(int32_t r1, int32_t r2);
void instr_0FFA_mem(int32_t addr, int32_t r);
void instr_660FFA(union reg128 source, int32_t r);
void instr_660FFA_reg(int32_t r1, int32_t r2);
void instr_660FFA_mem(int32_t addr, int32_t r);
void instr_0FFB(union reg64 source, int32_t r);
void instr_0FFB_reg(int32_t r1, int32_t r2);
void instr_0FFB_mem(int32_t addr, int32_t r);
void instr_660FFB(union reg128 source, int32_t r);
void instr_660FFB_reg(int32_t r1, int32_t r2);
void instr_660FFB_mem(int32_t addr, int32_t r);
void instr_0FFC(union reg64 source, int32_t r);
void instr_0FFC_reg(int32_t r1, int32_t r2);
void instr_0FFC_mem(int32_t addr, int32_t r);
void instr_660FFC(union reg128 source, int32_t r);
void instr_660FFC_reg(int32_t r1, int32_t r2);
void instr_660FFC_mem(int32_t addr, int32_t r);
void instr_0FFD(union reg64 source, int32_t r);
void instr_0FFD_reg(int32_t r1, int32_t r2);
void instr_0FFD_mem(int32_t addr, int32_t r);
void instr_660FFD(union reg128 source, int32_t r);
void instr_660FFD_reg(int32_t r1, int32_t r2);
void instr_660FFD_mem(int32_t addr, int32_t r);
void instr_0FFE(union reg64 source, int32_t r);
void instr_0FFE_reg(int32_t r1, int32_t r2);
void instr_0FFE_mem(int32_t addr, int32_t r);
void instr_660FFE(union reg128 source, int32_t r);
void instr_660FFE_reg(int32_t r1, int32_t r2);
void instr_660FFE_mem(int32_t addr, int32_t r);
void instr_0FFF(void);
void run_instruction0f_16(int32_t opcode);
void run_instruction0f_32(int32_t opcode);
jit_instr_flags jit_instruction0f_16(int32_t opcode);
jit_instr_flags jit_instruction0f_32(int32_t opcode);
